// ============================================================================
// Classes to handle a single message (e.g. find and get specific header-lines)
// Copyright (c) 1999, Juergen Haible.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
// ============================================================================

unit cArticle; // Class to handle a single message.

// ----------------------------------------------------------------------------
// Contains a class to handle a single message, which, for examples, allows
// direct access to specific header-lines.
// ----------------------------------------------------------------------------

interface

uses
  Classes;

type
  TArticle = class //old
  private
    fHeader: TStringList;
    fBody: String;
    function GetBody: String;
    function GetHeader(Const HdrName: String): String;
    function GetText: String;
    procedure SetBody(const Value: String);
    procedure SetHeader(Const HdrName: String; const Value: String);
    procedure SetText(const Value: String);
    function GetFullHeader: String;
    procedure SetFullHeader(const Value: String);
    Function FindHeader(Const HdrName: String; Var AktValue: String): Integer;
    function GetGMTDateDT: TDateTime;
    function GetHeaderLine(Const Line: Integer): String;
    function GetHeaderWOCRLF(const HdrName: String): String;
  public
    // Header
    Property FullHeader: String read GetFullHeader Write SetFullHeader;
    Function HeaderCount: Integer;
    Function FindHeaderIdx(Const HdrName: String; Var AktValue: String;
       Const StartIdx: Integer): Integer;
    Function HeaderAfterIdx(Const HdrName: String; Var Idx: Integer): String;
    Function HeaderExists(Const HdrName: String): boolean;
    Property Header[Const HdrName: String]: String read GetHeader Write SetHeader; default;
    Property HeaderWOCRLF[Const HdrName: String]: String read GetHeaderWOCRLF;
    Property Headerline[Const Line: Integer]: String read GetHeaderline;
    Procedure AddHeader (Const HdrName, Value: String); overload;
    Procedure AddHeader (Const Hdr: String); overload;
    Procedure AddHeaderFirst (Const HdrName, Value: String);
    procedure AddHeaderAtPos(const HdrName, Value: String; const p: Integer);
    Procedure AddHeaderAfterSpecialHeaders (Const HdrName, Value: String);
    Function DeleteHeader (Const HdrName: String): boolean;
    Function RenameHeader (Const OldHdrName, NewHdrName: String): boolean;
    Function QuotedHeader (Const QuoteWith: String): String;
    // Body
    Property FullBody: String read GetBody Write SetBody;
    // Alltogether
    Property Text: String read GetText Write SetText;
    Function QuotedText (Const QuoteWith: String): String;
    Function LoadFromFile(Const FileName: String): Integer;
    // Special functions
    function GetReceivedDT: TDateTime;
    function GetOurXHeader( Const XNam: String ) : String;
    procedure MimeDecodeHeaders;
    Function AsExportFormat: String;
    // Default
    constructor Create; overload;
    constructor Create (Const AktHeader, AktBody: String); overload;
    destructor Destroy; override;
  end;

  TMess = class
    private
      fHeaderLines: TStringList;
      fBodyText   : String;

      function  GetFullText: String;
      procedure SetFullText( const sText: String );
      function  GetHeaderText: String;
      procedure SetHeaderText( const sNewHdr: String );
      function  GetExportText: String;
      function  GetHeaderCountSL: Integer;
      function  GetHeaderLineSL( Index: Integer ): String;
      procedure SetHeaderLineSL( Index: Integer; const sNewValue: String );
      function  GetDateTimeGMT: TDateTime;
      function  GetBodyLines: Integer;

    public
      class function HeaderValueFromText( const ArtText, HdrName: String ): String;
      class function HasHeaderEnd( const ArtText: String ): Boolean;

      function IsValid: Boolean;
      property FullText: String read GetFullText write SetFullText;
      property HeaderLines: TStringList read fHeaderLines write fHeaderLines; // OZ - Ys enhancements
      property BodyText: String read fBodyText write fBodyText;
      property BodyLines: Integer read GetBodyLines;
      property HeaderText: String read GetHeaderText write SetHeaderText;
      property ExportText: String read GetExportText;

      property DateTimeGMT: TDateTime read GetDateTimeGMT;

      property HeaderLineCountSL: Integer read GetHeaderCountSL;
      property HeaderLineSL[ Index: Integer ]: String read GetHeaderLineSL write SetHeaderLineSL;
      function NextHeader( var Index: Integer; out Name, Value: String ): Boolean;

      function  AddHeaderSL( const sNewHeader: String ): Integer;
      function  AddHeaderML( const sNewHeader: String ): Integer;
      procedure InsertHeaderSL( Index: Integer; const sNewHeader: String );
      function  SetHeaderSL( sHdrName: String; const sHdrValue: String; sHdrNameSaveOld: String ): Integer;
      procedure DeleteHeaderML( Index: Integer );
      procedure DelHeaderML( const sHdrName: String );
      function  IndexOfHeader( const sHdrName: String; iAfterPos: Integer = -1 ): Integer;
      function  HeaderExists( const sHdrName: String ): Boolean;
      function  HeaderValueByIndexSL( Index: Integer ): String;
      function  HeaderValueByNameSL( const sHdrName: String ): String;
      function  HeaderValueByIndexML( Index: Integer ): String;
      function  HeaderValueByNameML( const sHdrName: String ): String;

      function  GetOurXHeader( const XNam: String ) : String;
      function  GetReceivedDT: TDateTime;
      function  GetHeaderLine( const HdrNam: String; var LineNo: Integer ) : String;

      function  LoadFromFile( const Filename: String ): Integer;
      function  SaveToFile( const Filename: String ): Integer;

      constructor Create;
      destructor Destroy; override;
  end;

function ParseMail( Const h: String ) : String;

implementation

uses SysUtils, Global, uTools, uEncoding, uDateTime;

function ParseMail( Const h: String ) : String;
var  s: String; pAt, p : Byte;
begin
   Result := '???@???';
   if h='' then exit;
   s := h;
   pAt := Pos( '@', s );
   if pAt>0 then begin
      repeat
         p := Pos( ' ', s );
         if p>0 then begin
            if p<pAt
               then s:=Copy(s,p+1,Length(s)-p)
               else s:=Copy(s,1,p-1);
         end;
      until p=0;
   end;

   if s>'' then Result:=s;
end;

Function CompleteHeaderName(Const s: String): String;
begin
   Result := Trim(s);
   If Result > '' then If Result[Length(Result)]<>':' then Result := Result + ':';
   Result := Result + ' '
end;

{ TArticle }

constructor TArticle.Create;
begin
   inherited Create;
   fHeader := TStringList.Create;
end;

constructor TArticle.Create(const AktHeader, AktBody: String);
begin
   Create;
   FullHeader := AktHeader;
   FullBody := AktBody
end;

destructor TArticle.Destroy;
begin
   fHeader.Free;
   inherited
end;

function TArticle.GetText: String;
begin
   Result := FullHeader;
   If FullBody > '' then Result := Result + #13#10 + FullBody
end;

procedure TArticle.SetText(const Value: String);
Var p: Integer;
begin
   p := Pos(#13#10#13#10, Value);
   If p > 0 then begin
      FullHeader := Copy(Value, 1, p-1);
      FullBody := Copy(Value, p+4, Length(Value))
   end else begin
      FullHeader := Value;
      FullBody := ''
   end
end;

function TArticle.GetBody: String;
begin
   Result := FBody
end;

procedure TArticle.SetBody(const Value: String);
begin
   fBody := Value
end;

function TArticle.GetFullHeader: String;
begin
   Result := fHeader.text
end;

procedure TArticle.SetFullHeader(const Value: String);
Var i: Integer; tl, tmax: Integer; Temp: String; c: Char; bEOL: Boolean;

  Procedure AddChar(Const c: Char);
  begin
     If tl >= tmax then begin
        tmax := tmax + 2048;
        SetLength(Temp, tMax)
     end;
     Inc(tl); Temp[tl] := c
  end;

  Procedure AddLine;
  Var s: String; p: Integer;
  begin
     If tl > 0 then begin
        s := Copy(Temp, 1, tl);
        If (Pos('mail from:', Lowercase(s))=0) and (Pos('rcpt to:', Lowercase(s))=0) then
           If (s[1]<>' ') and (s[1]<>^I) then begin
              p := Pos(':', s);
              If (p>0) and (p<Length(s)) then begin
                 Case s[p+1] of
                    ' ': ;
                    ^I : s[p+1] := ' ';
                    else Insert(' ', s, p+1)
                 end
              end
           end;
        fHeader.Add (s);
        tl := 0
     end
  end;

begin
   fHeader.Clear;
   bEOL := false;
   tmax := 2048;
   tl := 0;
   SetLength(Temp, tMax);
   For i := 1 to Length(Value) do begin
      c := Value[i];
      Case c of
         #13, #10: bEOL := true;
         ' ', ^I:
            begin
               If bEOL then begin
                  AddChar(#13);
                  AddChar(#10);
                  bEOL := false
               end;
               AddChar(c)
            end;
         else begin
            If bEOL then begin AddLine; bEOL := false end;
            AddChar(c)
         end
      end;
   end; // For
   AddLine
end;

procedure TArticle.AddHeaderAtPos(Const HdrName, Value: String; Const p: Integer);
Var s: String;
begin
   s := CompleteHeaderName(HdrName) + Value;
   If s = '' then Exit;
   If p < fHeader.Count then fHeader.Insert(p, s)
                        else fHeader.Add(s)
end;

procedure TArticle.AddHeader(const HdrName, Value: String);
Var s: String;
begin
   s := CompleteHeaderName(HdrName) + Value;
   AddHeader (s)
end;

Procedure TArticle.AddHeader (Const Hdr: String);
begin
   fHeader.Add (Hdr)
end;

procedure TArticle.AddHeaderFirst(const HdrName, Value: String);
begin
   AddHeaderAtPos(HdrName, Value, 0)
end;

procedure TArticle.AddHeaderAfterSpecialHeaders(const HdrName,
  Value: String);
Var i: Integer;
begin
   i := 0;
   While (i < fHeader.Count) and (Copy(fHeader[i], 1, 1)='!') do Inc(i);
   AddHeaderAtPos(HdrName, Value, i)
end;

Function TArticle.DeleteHeader(const HdrName: String): boolean;
Var AktValue: String; i: Integer;
begin
  i := FindHeader(HdrName, AktValue);
  Result := i >= 0;
  If Result then fHeader.Delete(i)
end;


function TArticle.GetHeader(Const HdrName: String): String;
Var AktValue: String;
begin
  FindHeader(HdrName, AktValue);
  Result := AktValue
end;

procedure TArticle.SetHeader(Const HdrName: String; const Value: String);
Var AktValue: String; i: Integer;
begin
  i := FindHeader(HdrName, AktValue);
  If i >= 0 then begin
     If AktValue <> Value then fHeader[i] := CompleteHeaderName(HdrName)+Value
  end else begin
     AddHeader (HdrName, Value)
  end
end;

function TArticle.RenameHeader(const OldHdrName,
  NewHdrName: String): boolean;
Var i: Integer; Value: String;
begin
   DeleteHeader (NewHdrName);
   i := FindHeader ( OldHdrName, Value );
   Result := i >= 0;
   If Result then fHeader[i] := CompleteHeaderName(NewHdrName) + Value
end;

function TArticle.HeaderCount: Integer;
begin
   Result := fHeader.Count
end;

function TArticle.HeaderAfterIdx(const HdrName: String;
  var Idx: Integer): String;
Var AktValue: String;
begin
  Idx := FindHeaderIdx (HdrName, AktValue, Idx);
  Result := AktValue
end;

function TArticle.QuotedHeader(const QuoteWith: String): String;
Var i: Integer;
begin
   With TStringList.Create do try
      Text := FullHeader;
      Result := '';
      For i := 0 to Count-1 do Result := Result + QuoteWith + Strings[i] + CRLF //JW QuotedHeader 13.05.02
   finally
      free
   end
end;

function TArticle.QuotedText(const QuoteWith: String): String;
Var i: Integer;
begin
   With TStringList.Create do try
      Text := Self.GetText;
      Result := '';
      For i := 0 to Count-1 do begin
         Result := Result + QuoteWith + Strings[i] + CRLF //JW QuotedHeader 13.05.02
      end
   finally
      free
   end
end;

Function TArticle.LoadFromFile(const FileName: String):integer;
Var Buf: String;
begin
   With TFileStream.Create ( FileName, fmOpenRead + fmShareDenyNone ) do try
      If Size > 0 then begin
      SetLength(Buf, Size);
      read (Buf[1], Size)
      end else begin
         Buf := ''
      end
   finally
      free
   end;
   Text := Buf;
   Result:=0;
end;

// ----------------------------------------------------------------

function TArticle.GetReceivedDT: TDateTime;
var  s: String;
begin
   Result := 0;
   s := GetOurXHeader( 'Received' );
   if s > '' then Result := TimeStampToDateTime(s);
   if Result = 0 then Result := GetGMTDateDT
end;

procedure TArticle.MimeDecodeHeaders;
Var i : Integer; s, name : String;
const FilteredHeaders: String =
       'from to cc subject organization reply-to resent-from resent-to resent-cc sender ';
begin
   for i := 0 to fHeader.Count-1 do begin
      s := fHeader[i];
      name := Copy( s, 1, pos( ':', s ) - 1 );
      if ( ( pos( Lowercase( name ) + ' ', FilteredHeaders ) > 0 ) or ( s > '' ) )
         then fHeader[i] := DecodeHeadervalue( s );
   end;
end;

Function TArticle.AsExportFormat: String;
var  DT        : TDateTime;
     hPosFrom, hPosCrlf: Integer; Line, TextBuf: String;
begin
   // introduction line ("From ???@??? Fri Sep 8 13:21:43 1995")
   DT := GetGMTDateDT;
   Result := 'From ' + ParseMail( Header['From:'] ) + ' '
              + copy( RFC_DAY_NAMES, DayOfWeek(DT)*3-2, 3 ) + ' '
              + copy( RFC_MONTH_NAMES, strtoint(FormatDateTime('mm',DT))*3-2, 3 ) + ' '
              + FormatDateTime( 'd hh:nn:ss yyyy', DT ) + #13#10;
   // article-text
   TextBuf := Text;
   repeat
      hPosFrom := Pos( #13#10 + 'From ', TextBuf );
      if hPosFrom > 0 then begin
         Result := Result + copy( TextBuf, 1, hPosFrom + 1);
         system.delete( TextBuf, 1, hPosFrom + 1);
         hPosCrlf := Pos( #13#10, TextBuf );
         if hPosCrlf > 0 then begin
           Line := copy(TextBuf, 1, hPosCrlf + 1);
           system.delete( TextBuf, 1, hPosCrlf + 1)
         end else begin
           Line := TextBuf;
           TextBuf := ''
         end;
         // quote lines with "From ..."
         if (copy(Line,1,5)='From ') then Line:='>'+Line;
         Result := Result + Line
      end
   until hPosFrom = 0;
   Result := Result + TextBuf
end;

function TArticle.GetOurXHeader(Const XNam: String): String;
Var h: String; i: Integer;
begin
   Result := '';
   h := Header[OUR_X_HEADER];
   if h='' then exit;
   i := Pos( LowerCase(XNam), LowerCase(h) );
   if i=0 then exit;
   System.Delete( h, 1, i+length(XNam) );
   i := Pos(' ',h);
   if i>0 then h:=copy(h,1,i-1);
   Result := h
end;


function TArticle.FindHeader(const HdrName: String; Var AktValue: String): Integer;
begin
   Result := FindHeaderIdx (HdrName, AktValue, -1)
end;

function TArticle.FindHeaderIdx(const HdrName: String; Var AktValue: String;
  const StartIdx: Integer): Integer;
Var v: String; vl, i: Integer;
begin
   v := Trim(LowerCase(CompleteHeaderName(HdrName)));
   vl := Length(v);
   Result := -1;
   AktValue := '';
   For i := StartIdx+1 to fHeader.Count-1 do begin
      If LowerCase(Copy(fHeader[i], 1, vl))=v then begin
         Result := i;
         AktValue := Copy(fHeader[i], vl+1, Length(fHeader[i])-vl);
         If AktValue > '' then begin
            If AktValue[1] IN [' ', #9] then Delete(Aktvalue, 1, 1);
         end;
         break
      end
   end
end;

function TArticle.GetGMTDateDT: TDateTime;
var s: String;
begin
   s := TrimWhSpace( Header['Date:']);
   if s = ''
      then Result := EncodeDate( 1980, 1, 1 )
      else Result := RfcDateTimeToDateTimeGMT( s )
end;

function TArticle.GetHeaderLine(Const Line: Integer): String;
begin
  Result := fHeader[Line];
end;

function TArticle.GetHeaderWOCRLF(const HdrName: String): String;
Var AktValue: String;
    i       : Integer;
begin
  FindHeader(HdrName, AktValue);
  for i := length(AktValue) downto 1 do
    if AktValue[i] in [#13, #10] then
      system.delete(AktValue,i,1);
  Result := AktValue
end;

function TArticle.HeaderExists(const HdrName: String): boolean;
Var AktValue: String;
begin
  Result := FindHeaderIdx (HdrName, AktValue, -1) >= 0
end;

// ---------------------------------------------------------------- TMess -----

class function TMess.HeaderValueFromText( const ArtText, HdrName: String ): String;
var  TempMess: TMess;
     i: Integer;
begin
   TempMess := TMess.Create;
   try
      i := Pos( CRLF+CRLF, ArtText );
      if i > 0 then TempMess.HeaderText := copy( ArtText, 1, i+1 );
      Result := TempMess.HeaderValueByNameSL( HdrName );
   finally TempMess.Free end;
end;

class function TMess.HasHeaderEnd( const ArtText: String ): Boolean;
begin
   Result := ( Pos( CRLF+CRLF, ArtText ) > 0 );
end;

function TMess.IsValid: Boolean;
begin
   Result := False;
   if fHeaderLines.Count = 0 then exit;
   Result := True;
end;

function TMess.GetFullText: String;
var  i: Integer;
begin
     Result := '';
     for i:=0 to fHeaderLines.Count-1 do begin
        Result := Result + fHeaderLines[i] + CRLF;
     end;
     Result := Result + CRLF;
     Result := Result + fBodyText;
end;

procedure TMess.SetFullText( const sText: String );
var  i: Integer;
begin
     fHeaderLines.Clear;
     fBodyText := '';

     i := Pos( CRLF+CRLF, sText );
     if i=0 then begin
        fHeaderLines.Text := sText;
     end else begin
        fHeaderLines.Text := copy( sText, 1, i+1 );
        fBodyText := sText;
        System.Delete( fBodyText, 1, i+3 );
     end;
end;

function TMess.GetHeaderText: String;
var  i: Integer;
begin
     Result := '';
     for i:=0 to fHeaderLines.Count-1 do begin
        Result := Result + fHeaderLines[i] + CRLF;
     end;
end;

procedure TMess.SetHeaderText( const sNewHdr: String );
begin
     fHeaderLines.Clear;
     fHeaderLines.Text := sNewHdr;
end;

function TMess.GetHeaderCountSL: Integer;
begin
     Result := fHeaderLines.Count;
end;

function TMess.GetHeaderLineSL( Index: Integer ): String;
begin
     Result := fHeaderLines[ Index ];
end;

procedure TMess.SetHeaderLineSL( Index: Integer; const sNewValue: String );
begin
     fHeaderLines[ Index ] := sNewValue;
end;

function TMess.AddHeaderSL( const sNewHeader: String ): Integer;
begin
     Result := fHeaderLines.Add( sNewHeader );
end;

function TMess.AddHeaderML( const sNewHeader: String ): Integer;
var  TS: TStringList;
     i : Integer;
begin
     TS := TStringList.Create;
     TS.Text := sNewHeader;
     Result := fHeaderLines.Add( TS[0] );
     for i:=1 to TS.Count-1 do fHeaderLines.Add( TS[i] );
     TS.Free;
end;

procedure TMess.InsertHeaderSL( Index: Integer; const sNewHeader: String );
begin
     fHeaderLines.Insert( Index, sNewHeader );
end;

function TMess.SetHeaderSL( sHdrName: String; const sHdrValue: String; sHdrNameSaveOld: String ): Integer;
var  Index : Integer;
     sOldVal: String;
begin
     if sHdrName[length(sHdrName)]<>':' then sHdrName:=sHdrName+':';
     Index := IndexOfHeader( sHdrName, -1 );

     if Index>=0 then begin
        sOldVal := HeaderValueByIndexML( Index );
        DeleteHeaderML( Index );
        if sHdrNameSaveOld<>'' then begin
           if sHdrNameSaveOld[length(sHdrNameSaveOld)]<>':' then sHdrNameSaveOld:=sHdrNameSaveOld+':';
           AddHeaderML( sHdrNameSaveOld + ' ' + sOldVal );
        end;
     end;

     Result := AddHeaderSL( sHdrName + ' ' + sHdrValue );
end;

procedure TMess.DeleteHeaderML( Index: Integer );
begin
     fHeaderLines.Delete( Index );
     repeat
        if Index>=fHeaderLines.Count then break;
        if fHeaderLines[Index] = '' then break;
        if not( fHeaderLines[Index][1] in [#9,' '] ) then break;
        fHeaderLines.Delete( Index );
     until False;
end;

procedure TMess.DelHeaderML( const sHdrName: String );
var  Index : Integer;
begin
     Index := IndexOfHeader( sHdrName, -1 );
     if Index>=0 then DeleteHeaderML( Index );
end;

function TMess.IndexOfHeader( const sHdrName: String; iAfterPos: Integer = -1 ): Integer;
var  i, k: Integer;
     h, s: String;
begin
     Result := -1;

     h := LowerCase( TrimWhSpace( sHdrName ) );
     If h = '' then exit;
     if h[length(h)]=':' then System.Delete(h,length(h),1);

     i := iAfterPos + 1;
     while i<fHeaderLines.Count do begin
        s := fHeaderLines[i];
        if s='' then break;
        if not( s[1] in [#9,' '] ) then begin
           k := Pos( ':', s );
           if k>0 then begin
              s := LowerCase( copy( s, 1, k-1 ) );
              if s=h then begin Result:=i; break; end;
           end;
        end;
        inc( i );
     end;
end;

function TMess.HeaderExists( const sHdrName: String ): Boolean;
begin
   Result := ( IndexOfHeader(sHdrName) >= 0 );
end;

function TMess.NextHeader( var Index: Integer; out Name, Value: String ): Boolean;
var  i: Integer;
     s: String;
begin
   Result := False;
   if Index >= fHeaderLines.Count then exit;

   Result := True;
   s := fHeaderLines[Index];
   i := Pos( ':', s );
   Name := copy( s, 1, i-1 );
   System.Delete( s, 1, i );
   if copy( s, 1, 1 ) = ' ' then System.Delete( s, 1, 1 );
   Value := s;
   inc( Index );

   while Index < fHeaderLines.Count do begin
      if fHeaderLines[Index] = '' then break;
      if not( fHeaderLines[Index][1] in [#9,' '] ) then break;
      Value := Value + fHeaderLines[Index];
      inc( Index );
   end;
end;

function TMess.HeaderValueByIndexSL( Index: Integer ): String;
var  i: Integer;
begin
     Result := fHeaderLines[Index];
     i := Pos( ':', Result );
     System.Delete( Result, 1, i );
     if copy( Result, 1, 1 ) = ' ' then System.Delete( Result, 1, 1 );

     while Index+1<fHeaderLines.Count do begin
        inc( Index );
        if fHeaderLines[Index] = '' then break;
        if not( fHeaderLines[Index][1] in [#9,' '] ) then break;
        Result := Result + fHeaderLines[Index];
     end;
end;

function TMess.HeaderValueByNameSL( const sHdrName: String ): String;
var  Index: Integer;
begin
     Index := IndexOfHeader( sHdrName, -1 );
     if Index>=0 then Result := HeaderValueByIndexSL( Index )
                 else Result := '';
end;

function TMess.HeaderValueByIndexML( Index: Integer ): String;
var  i: Integer;
begin
     Result := fHeaderLines[Index];
     i := Pos( ':', Result );
     System.Delete( Result, 1, i );
     if copy( Result, 1, 1 ) = ' ' then System.Delete( Result, 1, 1 );

     while Index+1<fHeaderLines.Count do begin
        inc( Index );
        if fHeaderLines[Index] = '' then break;
        if not( fHeaderLines[Index][1] in [#9,' '] ) then break;
        Result := Result + #13#10 + fHeaderLines[Index];
     end;
end;

function TMess.HeaderValueByNameML( const sHdrName: String ): String;
var  Index: Integer;
begin
     Index := IndexOfHeader( sHdrName, -1 );
     if Index>=0 then Result := HeaderValueByIndexML( Index )
                 else Result := '';
end;

function TMess.GetBodyLines: Integer;
var  SL: TStringList;
begin
   SL := TStringList.Create;
   try
      SL.Text := BodyText;
      Result := SL.Count;
   finally
      SL.Free;
   end;
end;

function TMess.GetDateTimeGMT: TDateTime;
begin
   Result := RfcDateTimeToDateTimeGMT( HeaderValueByNameSL( 'Date:' ) );
end;

function TMess.GetOurXHeader( const XNam: String ) : String;
var  h: String;
     i: Integer;
begin
     Result := '';

     h := HeaderValueByNameSL('X-Hamster-Info:');
     if h='' then exit;

     i := Pos( LowerCase(XNam), LowerCase(h) );
     if i=0 then exit;

     System.Delete( h, 1, i+length(XNam) );
     i := Pos(' ',h);
     if i>0 then h:=copy(h,1,i-1);

     Result := h;
end;

function TMess.GetReceivedDT: TDateTime;
var  s: String;
begin
     Result := 0;
     s := GetOurXHeader( 'Received' );
     if s<>''    then Result := TimeStampToDateTime(s);
     if Result=0 then Result := DateTimeGMTToLocal( GetDateTimeGMT );
end;

function TMess.GetHeaderLine( const HdrNam: String; var LineNo: Integer ) : String;
begin
   LineNo := IndexOfHeader( HdrNam );
   if LineNo < 0 then Result := ''
                 else Result := HeaderValueByIndexSL( LineNo );
end;

function TMess.GetExportText: String;
const DOW = 'SunMonTueWedThuFriSat';
      MOY = 'JanFebMayAprMayJunJulAugSepOctNovDec';

     function IntroductionLine: String;
     var  DT: TDateTime;
     begin
          // introduction line ("From ???@??? Fri Sep 8 13:21:43 1995")
          DT := DateTimeGMT;
          Result := 'From' + ' '
                  + ParseMail( HeaderValueByNameSL('From:') ) + ' '
                  + copy( DOW, DayOfWeek(DT)*3-2, 3 ) + ' '
                  + copy( MOY, strtoint(FormatDateTime('mm',DT))*3-2, 3 ) + ' '
                  + FormatDateTime( 'd hh:nn:ss yyyy', DT );
     end;

var  SL: TStringList;
     i : Integer;
begin
   SL := TStringList.Create;
   try
      SL.Text := FullText;
      if SL.Count > 0 then begin
         if copy( SL[0], 1, 5 ) <> 'From ' then begin

            SL.Insert( 0, IntroductionLine ); // add 'From ...' line

            for i:=1 to SL.Count-1 do begin // quote lines with leading "From "
               if ( copy( SL[i], 1, 5 ) = 'From ' ) (* and
                  ( Pos( '@', SL[i] ) > 0 ) *) then begin
                  SL[i] := '>' + SL[i];
               end;
            end;

         end;
      end;

      Result := SL.Text;

   finally
      SL.Free;
   end;
end;

function TMess.LoadFromFile( const Filename: String ): Integer;
var  Data: String;
     FS  : TFileStream;
begin
   Result := -1; // file not found
   Data   := '';

   try
      if FileExists( Filename ) then begin
         FS := TFileStream.Create( Filename, fmOpenRead or fmShareDenyNone );
         try
            SetLength( Data, FS.Size );
            FS.Read( Data[1], length(Data) );
            Result := 0; // ok
         finally FS.Free end;
      end;
   except
      Result := -2; // error
      Data   := '';
   end;

   FullText := Data;
end;

function TMess.SaveToFile( const Filename: String ): Integer;
var  Data: String;
     FS  : TFileStream;
begin
   Data   := FullText;
   Result := 0; // ok

   try
      FS := TFileStream.Create( Filename, fmCreate );
      try
         if length(Data) > 0 then FS.Write( Data[1], length(Data) );
      finally FS.Free end;
   except
      Result := -2; // error
      Data   := '';
   end;
end;

constructor TMess.Create;
begin
     inherited Create;
     fHeaderLines := TStringList.Create;
     fBodyText    := '';
end;

destructor TMess.Destroy;
begin
     if Assigned(fHeaderLines) then fHeaderLines.Free;
     inherited Destroy;
end;

end.
