// ============================================================================
// Hamster, a free news- and mailserver for personal, family and workgroup use.
// Copyright (c) 1999, Juergen Haible.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
// ============================================================================

unit cFiltersMail;

interface

uses SysUtils, Classes, Global, cArticle, cFiltersBase, cPCRE;

type
  TFilterPatternMail = class( TFilterPatternBase )
    FilterHeader: String;
    ExtractField: boolean;
  end;

  TFilterLineMail = class( TFilterLineBase )
    protected
      function MatchesAccounts( Hdr: TArticle; var ResultAccounts: String ): Boolean;
      function TestHeaderValue( Const HdrVal: String; Hdr: TArticle; RE: TPCRE ): Boolean;
    public
      IsFinal      : Boolean; // '='?
      DoAllHeaders : Boolean; // '*'?
      WithHeaderName: Boolean; // '^'?
      ActionID     : Integer;
      ActionPars   : String;
      DefaultField : String;
      Expire       : TDateTime;
      function  MatchesMailHeaders( RE: TPCRE; Hdr: TArticle; var ResultAccounts: String ): Boolean;
      function  SetFilterLine( FilterLine: String ): Boolean; override;
      function  AsString: String; override;
  end;

  TFiltersMail = class( TFiltersBase )
    public
      TOP_makes_sense  : Boolean;
      Filter_bytes_only: Boolean;
      Procedure Test;
      function  LinesAdd( Const LineText: String ): Integer; override;
      function  IsFilterLine( Const Line: String ): Boolean; override;
      procedure SelectSections( Const SectionIdentifier: String ); override;
      procedure FilterMail( MailHeaders      : String;
                                   DefaultUser      : String;
                                   var ResultIgnore : Boolean;
                                   var ResultKillIt : Boolean;
                                   ResultNotifys    : TStrings;
                                   ResultUsers      : TStrings;
                                   ResultGroups     : TStrings;
                                   var NotifyReason : String;
                                   var ResultIMAPFolder : String);
      Procedure Purge;
  end;

implementation

uses uTools, cAccount, Config, uEncoding, cLogFile, cStdForm;

const
  MLFLT_ACTION_INVALID     =  0;
  MLFLT_ACTION_LOAD        =  1;
  MLFLT_ACTION_KILL        =  2;
  MLFLT_ACTION_IGNORE      =  3;
  MLFLT_ACTION_NOTIFY      =  4;
  MLFLT_ACTION_NOTIFYOFF   =  5;  // JAWO 21.09.01 (Notify none)
  MLFLT_ACTION_DEFAULT     = 10;
  MLFLT_ACTION_SET         = 11;
  MLFLT_ACTION_ADD         = 12;
  MLFLT_ACTION_DEL         = 13;
  MLFLT_ACTION_ADDACCOUNTS = 14;
  MLFLT_ACTION_POSTTO      = 15;
  MLFLT_ACTION_ADDDEFAULT  = 16;
  MLFLT_ACTION_FOLDER      = 17; //HSR //IMAP-Folder 08/02

const
  ANY_SENDER        = 'Any-Sender:';
  ANY_SENDER_MAX    = 7;
  ANY_SENDER_LIST   : array[0..ANY_SENDER_MAX] of String =
                      ( 'From:', 'Apparently-From:', 'Sender:', 'Reply-To:',
                        'X-Sender', 'Envelope-From:', 'X-Envelope-From:',
                        'Return-Path:' );

  ANY_RECIPIENT     = 'Any-Recipient:';
  ANY_RECIPIENT_MAX = 5;
  ANY_RECIPIENT_LIST: array[0..ANY_RECIPIENT_MAX] of String =
                      ( 'To:', 'Apparently-To:', 'CC:', 'BCC:',
                        'Envelope-To:', 'X-Envelope-To:' );

  ANY_HEADER        = 'Any-Header:';

procedure ExpandFilterGroups( HdrName: String; TS: TStrings; Hdr: TArticle );
var  i: Integer;
     s: String;
begin
   If HdrName = '' then Exit;
   if HdrName[length(HdrName)]<>':' then HdrName:=HdrName+':';
   if CompareText( ANY_SENDER, HdrName )=0 then begin
      for i:=0 to ANY_SENDER_MAX do TS.Add( ANY_SENDER_LIST[i] );
   end else
   if CompareText( ANY_RECIPIENT, HdrName )=0 then begin
      for i:=0 to ANY_RECIPIENT_MAX do TS.Add( ANY_RECIPIENT_LIST[i] );
   end else
   if CompareText( ANY_HEADER, HdrName )=0 then begin
     for i:=0 to Hdr.HeaderCount - 1 do begin
        s := Hdr.Headerline[i];
        if Pos( ': ', s ) > 0 then TS.Add( Copy( s, 1, Pos( ': ', s ) ) );
     end;
     exit;
   end else
   begin
      TS.Add( HdrName )
   end
end;

// ------------------------------------------------------ TFilterLineMail -----

function TFilterLineMail.MatchesAccounts( Hdr: TArticle; var ResultAccounts: String ): Boolean;
var  TestHdrs, MailAddrs: TStringList;
     TestHdr, DefStr, s: String;
     CurrHdr, LastPos, UID, LTY, i: Integer;
     bLocal: Boolean;

     procedure AddMailAddress( TestStr: String );
     var  at: Integer;
     begin
          TestStr := TrimWhSpace( TestStr );
          if length(TestStr)<3 then exit;
          if PosWhSpace( TestStr )>0 then exit;

          if TestStr[1]='<' then System.Delete( TestStr, 1, 1 );
          if TestStr[length(TestStr)]='>' then System.Delete( TestStr, length(TestStr), 1 );

          at := Pos( '@', TestStr );
          if (at<2) or (at=length(TestStr)) then exit;

          MailAddrs.Add( TestStr );
     end;

     procedure AddMailAddresses( HdrVal: String );
     // add all strings which look like mail-addresses to accounts-testlist
     var  p: Integer;
          addr: String;
     begin
          addr := '';
          p    := 1;

          while p<=length(HdrVal) do begin
             if HdrVal[p] in [#9,' ','<','>'] then begin
                AddMailAddress( addr );
                addr := '';
             end else begin
                addr := addr + HdrVal[p];
             end;

             inc( p );
          end;

          AddMailAddress( addr );
     end;
begin
   TestHdrs := TStringList.Create;
   TestHdrs.Sorted := True;
   TestHdrs.Duplicates := dupIgnore;
   MailAddrs := TStringList.Create;
   try
     ExpandFilterGroups( DefaultField, TestHdrs, Hdr );
     MailAddrs.Sorted := True;
     MailAddrs.Duplicates := dupIgnore;
     for CurrHdr:=0 to TestHdrs.Count-1 do begin
        TestHdr := TestHdrs[CurrHdr];
        LastPos := -1;
        repeat
           DefStr := Hdr.HeaderAfterIdx( TestHdr, LastPos );
           If DefStr > '' then begin
              If DoMimeDecode then DefStr:=DecodeHeadervalue(DefStr);
              AddMailAddresses( DefStr )
           end
        until (DefStr = '') or (not DoAllHeaders);
     end;

     for i:=0 to MailAddrs.Count-1 do begin
        if MailAddrs[i]<>'' then begin
           if CfgAccounts.IsLocalMailbox( MailAddrs[i], False, UID, LTY, bLocal ) then begin
              s := CfgAccounts.Value[ UID, ACTP_USERNAME ];
              if ResultAccounts<>'' then ResultAccounts:=ResultAccounts+',';
              ResultAccounts := ResultAccounts + s;
           end else begin
              Log( LOGID_DEBUG, 'MailFilter.addaccounts: "' + MailAddrs[i] + '" unknown.' );
           end;
        end;
     end;

     if ResultAccounts<>'' then Log( LOGID_DEBUG, 'MailFilter.addaccounts.final: "' + ResultAccounts + '"' );
     Result := ( ResultAccounts <> '' )
   finally
     MailAddrs.Free;
     TestHdrs.Free;
   end
end;

function TFilterLineMail.TestHeaderValue( Const HdrVal: String; Hdr: TArticle; RE: TPCRE ): Boolean;
var  Matches, NeedOneOf, HaveOneOf: Boolean;
     PatNo: Integer;
     TestStr: String;
     Pat: TFilterPatternMail;
     Pattern: String;
begin
     Result := True;

     NeedOneOf := False;
     HaveOneOf := False;
     Matches   := False;

     for PatNo:=0 to PatternCount-1 do begin
        Pat := TFilterPatternMail( PatternItem[ PatNo ] );

        If Pat.ExtractField
           then Pattern := Hdr[Pat.Pattern]
           else Pattern := Pat.Pattern;
        If Pattern = '' then continue;

        if CompareText( Pat.FilterHeader, DefaultField )=0 then begin
           TestStr := HdrVal;
        end else begin
           TestStr := Hdr[Pat.FilterHeader];
           if (TestStr>'') and DoMimeDecode then TestStr:=DecodeHeadervalue(TestStr);
        end;

        if (Pat.SelectType<>' ') or not(HaveOneOf) then begin
           if Pat.IsRegex then begin
              try
                 RE.OptCompile := PCRE_CASELESS;
                 Matches := RE.Match( PChar(Pattern), PChar(TestStr) );
              except
                 on E: Exception do begin
                    Log( LOGID_ERROR, 'Regex-error in {' + Pattern + '}:' + E.Message );
                    Matches := False;
                 end;
              end;
           end else begin
              Matches := MatchSimple( TestStr, Pattern );
           end;

           case Pat.SelectType of
              '+': if not Matches then begin Result:=False; break; end;
              '-': if Matches     then begin Result:=False; break; end;
              ' ': begin
                      NeedOneOf := True;
                      if Matches then HaveOneOf:=True;
                   end;
           end;
        end;
     end;

     if NeedOneOf and not HaveOneOf then Result:=False;
end;

function TFilterLineMail.MatchesMailHeaders( RE: TPCRE; Hdr: TArticle;
   var ResultAccounts: String ): Boolean;
var  TestHdr, DefStr, HdrValue: String;
     CurrHdr, LastPos: Integer;
     TestHdrs: TStringList;
begin
   ResultAccounts := '';
   Case ActionID  of
      // test given header-name[s] for known mail-accounts
      MLFLT_ACTION_ADDACCOUNTS:
         Result := MatchesAccounts( Hdr, ResultAccounts );
      // test given patterns
      MLFLT_ACTION_SET, MLFLT_ACTION_ADD, MLFLT_ACTION_DEL,
         MLFLT_ACTION_POSTTO, MLFLT_ACTION_LOAD, MLFLT_ACTION_KILL,
         MLFLT_ACTION_IGNORE:
      begin
         Result := Defaultfield = '';
         TestHdrs := TStringList.Create;
         TestHdrs.Sorted := True;
         TestHdrs.Duplicates := dupIgnore;
         try
            ExpandFilterGroups( DefaultField, TestHdrs, Hdr );
            for CurrHdr:=0 to TestHdrs.Count-1 do begin
               TestHdr := TestHdrs[CurrHdr];
               LastPos := -1;
               repeat
                  LastPos := Hdr.FindHeaderIdx( TestHdr, HdrValue, LastPos );
                  if LastPos<0 then break;
                  if WithHeaderName
                     then DefStr := Hdr.HeaderLine[ LastPos ]
                     else DefStr := HdrValue;
                  If DoMimeDecode then DefStr:=DecodeHeadervalue(DefStr);
                  // test given patterns
                  Result := TestHeaderValue( DefStr, Hdr, RE ); // test patterns
               until Result or (DefStr = '') or (Not DoAllHeaders);
               if Result then break;
            end;
         finally
            TestHdrs.Free;
         end
      end;
      // Some Actions are handled outside
      { MLFLT_ACTION_NOTIFY, MLFLT_ACTION_NOTIFYOFF,
        MLFLT_ACTION_DEFAULT, MLFLT_ACTION_ADDDEFAULT }
      ELSE Result := true;
   end
end;

Type TVTyp = (vtText, vtField, vtRegex);

function TFilterLineMail.SetFilterLine( FilterLine: String ): Boolean;
var  i, j, k: Integer;
     s, h: String;
     SelectType: Char;
     VTyp: TVTyp;
     FilterHeader: String;
     Pattern: String;
     cEnd: Char;
     pat: TFilterPatternMail;
     bError: Boolean;
begin
     Result := False;
     LastSetError := 'invalid line';

     Clear;
     IsFinal        := False;
     ActionID       := MLFLT_ACTION_INVALID;
     ActionPars     := '';
     DoMimeDecode   := False;
     DoAllHeaders   := False;
     WithHeaderName := False;
     DefaultField   := '';

     FilterLine := TrimWhSpace( FilterLine );
     if FilterLine='' then exit;

     // Final-marker
     if FilterLine[1]='=' then begin
        IsFinal:=True;
        System.Delete( FilterLine, 1, 1 );
     end;

     // action
     i := PosWhSpace( FilterLine );
     // This part filters blanks in the first section, if there are brackets
     If Length(FilterLine) > i then begin
        j := 0;
        For k := i+1 to Length(Filterline) do begin
           Case Filterline[k] of
              ' ', ^I: ;
              '(': begin j := k; break end;
              else break
           end
        end;
        If j = 0 then begin
           For k := 1 to i-1 do If Filterline[k] = '(' then begin j := k; break end
        end;
        If j > 0 then begin
           i := Pos(')', Filterline);
           If i = 0 then begin LastSetError:='missing ")" in "'+Filterline+'"'; exit end;
           For k := i-1 downto 1 do begin
              If (Filterline[k]=' ') or (Filterline[k]=^I) then Delete(Filterline, k, 1)
           end;
           i := PosWhSpace( FilterLine );
        end
     end;
     if i=0 then i:=length(FilterLine)+1;
     if i<3 then begin LastSetError:='missing filter-action'; exit; end;
     s := copy( FilterLine, 1, i-1 );
     System.Delete( FilterLine, 1, i );
     FilterLine := TrimWhSpace( FilterLine );

     i := Pos( '(', s );
     if i=0 then begin
        h := s;
        s := '';
     end else begin
        h := copy( s, 1, i-1 );
        System.Delete( s, 1, i );
      if copy(s,length(s),1)<>')' then begin
         LastSetError:='missing ")" in "'+h+'('+s+'"';
         exit;
      end;
        System.Delete( s, length(s), 1 );
     end;
     ActionPars := TrimWhSpace(s);

     h := LowerCase( TrimWhSpace( h ) );
     if      h='default'     then ActionID:=MLFLT_ACTION_DEFAULT
     else if h='set'         then ActionID:=MLFLT_ACTION_SET
     else if h='add'         then ActionID:=MLFLT_ACTION_ADD
     else if h='del'         then ActionID:=MLFLT_ACTION_DEL
     else if h='addaccounts' then ActionID:=MLFLT_ACTION_ADDACCOUNTS
     else if h='adddefault'  then ActionID:=MLFLT_ACTION_ADDDEFAULT
     else if h='postto'      then ActionID:=MLFLT_ACTION_POSTTO
     else if h='load'        then ActionID:=MLFLT_ACTION_LOAD
     else if h='kill'        then ActionID:=MLFLT_ACTION_KILL
     else if h='ignore'      then ActionID:=MLFLT_ACTION_IGNORE
     else if h='notify'      then ActionID:=MLFLT_ACTION_NOTIFY
     else if h='notifyoff'   then ActionID:=MLFLT_ACTION_NOTIFYOFF   // JAWO 21.09.01 (Notify none)
     else if h='folder'      then ActionID:=MLFLT_ACTION_FOLDER
     else                         ActionID:=MLFLT_ACTION_INVALID;
     if ActionID=MLFLT_ACTION_INVALID then begin
        LastSetError := 'Invalid action "' + h + '"';
        exit;
     end;

   Filterline := TrimWhSpace( FilterLine );
   If CheckExpire ( Filterline, LastSetError, Expire, bError ) then Filterline := TrimWhSpace( FilterLine );
   If bError then Exit;

     // check filter-params
     if not( ActionID in [MLFLT_ACTION_DEFAULT, MLFLT_ACTION_NOTIFY,
           MLFLT_ACTION_NOTIFYOFF, MLFLT_ACTION_ADDDEFAULT ] ) then // JAWO 21.09.01 (Notify none)
     begin

        s := TrimWhSpace( FilterLine );
        If (s = '') or (s[1] in ['#',';']) then begin
           DefaultField := ''
        end else begin
           // Default-Field with optional MIME-Decode
           i := PosWhSpace( FilterLine );
           if i = 0 then i:=length(FilterLine)+1;
           k := Pos( ':', FilterLine );
           If (k > 0) and (k < i) then i := k+1;
           if i<2 then begin LastSetError:='missing: [~][*][^]default-field'; exit; end;
           s := copy( FilterLine, 1, i-1 );
           System.Delete( FilterLine, 1, i-1 );
           FilterLine := TrimWhSpace( FilterLine );
           while (s>'') and (s[1] in ['~','*','^']) do begin
              if s[1]='~' then DoMimeDecode := True;
              if s[1]='*' then DoAllHeaders := True;
              if s[1]='^' then WithHeaderName := True;
              System.Delete( s, 1, 1 );
           end;
           if s = '' then begin LastSetError:='missing: [~][*][^]default-field'; exit; end;
           if s[length(s)]<>':' then s:=s+':';
           DefaultField := s;

         If CheckExpire ( s, LastSetError, Expire, bError ) then s := TrimWhSpace( s );
         If bError then Exit;

           if ActionID<>MLFLT_ACTION_ADDACCOUNTS then begin

              // One or more patterns of the following forms:
              //    ['+'|'-'] ['@' fieldname ':'] '{' regex-pattern '}'
              //    ['+'|'-'] ['@' fieldname ':'] '"' simple-pattern '"'
              //    ['+'|'-'] ['@' fieldname ':'] simple-pattern without WHSP

              while FilterLine>'' do begin

               Filterline := TrimWhSpace( FilterLine );
               If CheckExpire ( Filterline, LastSetError, Expire, bError ) then Filterline := TrimWhSpace( FilterLine );
               If bError then Exit;

               If FilterLine>'' then begin

                 SelectType := ' ';
                 if FilterLine[1] in ['+','-'] then begin
                    SelectType := FilterLine[1];
                    System.Delete( FilterLine, 1, 1 );
                 end;

                 FilterHeader := DefaultField;
                 if FilterLine[1]='@' then begin
                    i := Pos( ':', FilterLine );
                    if i<3 then begin LastSetError:='missing: "fieldname:" on "@"'; exit end;
                    s := copy( FilterLine, 2, i-2 );
                    System.Delete( FilterLine, 1, i );
                    if s[length(s)]<>':' then s:=s+':';
                    FilterHeader := s;
                 end;

                 if FilterLine='' then begin LastSetError:='missing: pattern'; exit; end;
                 Pattern := '';

                 if FilterLine[1]='{' then begin
                    VTyp := vtRegex;
                    System.Delete( FilterLine, 1, 1 );
                    k := 1;
                    while FilterLine<>'' do begin
                       if FilterLine[1]='{' then inc(k);
                       if FilterLine[1]='}' then begin
                          dec(k);
                          if k=0 then break;
                       end;
                       Pattern := Pattern + FilterLine[1];
                       System.Delete( FilterLine, 1, 1 );
                    end;
                    if copy(FilterLine,1,1)='}' then System.Delete( FilterLine, 1, 1 );
                 end else begin
                    if FilterLine[1]='$' then begin
                       VTyp := vtField;
                       cEnd := '$';
                       System.Delete( FilterLine, 1, 1 );
                       i := Pos( cEnd, FilterLine )
                    end else begin
                       VTyp := vtText;
                       if FilterLine[1]='"' then begin
                          cEnd := '"';
                          System.Delete( FilterLine, 1, 1 );
                          i := Pos( cEnd, FilterLine )
                       end else begin
                          cEnd := #32;
                           if (FilterLine[1] = '%') then begin
                              While (Length(FilterLine)>2) and ((FilterLine[3] = #32) or (FilterLine[3] = #9)) do begin
                                 System.Delete( FilterLine, 3, 1 )
                              end
                           end;
                          i := PosWhSpace( FilterLine )
                       end
                    end;
                    if (i<=0) then i := length(FilterLine) + 1;
                    Pattern := copy( FilterLine, 1, i-1 );
                    System.Delete( FilterLine, 1, i-1 );
                    if copy(FilterLine,1,1)=cEnd then System.Delete( FilterLine, 1, 1 )
                 end;

                  if Pattern='' then begin
                     LastSetError:='missing: pattern/-delimiter'; exit;
                  end;

                 pat := TFilterPatternMail.Create;
                 pat.SelectType   := SelectType;
                 pat.IsRegex      := ( VTyp = vtRegEx );
                 pat.ExtractField := ( VTyp = vtField );
                 pat.Pattern      := Pattern;
                 pat.FilterHeader := FilterHeader;
                 PatternAdd( pat );

                 FilterLine := TrimWhSpace( FilterLine );
                 if FilterLine<>'' then begin
                    if FilterLine[1] in ['#',';'] then FilterLine:=''; // rest of line is [valid] comment
                 end
              end
           end;
            If fFilterPatterns.Count = 0 then begin
               LastSetError:='missing: any pattern after default-field';
               Exit;
            end

         end
        end
     end;

     LastSetError := '';
     Result := True;
end;

function TFilterLineMail.AsString: String;
var  Pat  : TFilterPatternMail;
     PatNo: Integer;
begin
     Result := '';

     if IsFinal then Result:=Result+'=';

     case ActionID of
        MLFLT_ACTION_DEFAULT:     Result:=Result+'default';
        MLFLT_ACTION_SET:         Result:=Result+'set';
        MLFLT_ACTION_ADD:         Result:=Result+'add';
        MLFLT_ACTION_DEL:         Result:=Result+'del';
        MLFLT_ACTION_ADDACCOUNTS: Result:=Result+'addaccounts';
        MLFLT_ACTION_ADDDEFAULT:  Result:=Result+'adddefault'; 
        MLFLT_ACTION_POSTTO:      Result:=Result+'postto';
        MLFLT_ACTION_LOAD:        Result:=Result+'load';
        MLFLT_ACTION_KILL:        Result:=Result+'kill';
        MLFLT_ACTION_IGNORE:      Result:=Result+'ignore';
        MLFLT_ACTION_NOTIFY:      Result:=Result+'notify';
        MLFLT_ACTION_NOTIFYOFF:   Result:=Result+'notifyoff';   // JAWO 21.09.01 (Notify none)
        MLFLT_ACTION_FOLDER:      Result:=Result+'folder'; //HSR //IMAP-Folder 08/02
        else                      Result:=Result+'???';
     end;
     Result := Result + '(' + ActionPars + ')';

     if DefaultField<>'' then begin
        Result := Result + ' ';
        if DoMimeDecode then Result:=Result+'~';
        if DoAllHeaders then Result:=Result+'*';
        if WithHeaderName then Result:=Result+'^';
        Result := Result + DefaultField;

        for PatNo:=0 to PatternCount-1 do begin
           Pat := TFilterPatternMail( PatternItem[ PatNo ] );
           Result := Result + ' ';
           if Pat.SelectType in ['+','-'] then Result:=Result+Pat.SelectType;
           if CompareText( Pat.FilterHeader, DefaultField )<>0 then begin
              Result := Result + '@' + Pat.FilterHeader + ':';
           end;
           if Pat.IsRegEx then begin
              Result := Result + '{' + Pat.Pattern + '}';
           end else begin
              if Pos(' ',Pat.Pattern)>0 then begin
                 Result := Result + '"' + Pat.Pattern + '"';
              end else begin
                 Result := Result + Pat.Pattern;
              end;
           end;
        end;
     end;
end;

// --------------------------------------------------------- TFiltersMail -----

function TFiltersMail.LinesAdd( Const LineText: String ): Integer;
var  lin: TFilterLineMail;
begin
     lin := TFilterLineMail.Create;
     if lin.SetFilterLine( LineText ) then begin
        Result := fFilterLines.Add( lin );
     end else begin
        Log( LOGID_WARN, 'Filter-line ignored: ' + LineText );
        Log( LOGID_WARN, 'Reason: ' + lin.LastSetError );
        lin.Free;
        Result := -1;
     end;
end;

function TFiltersMail.IsFilterLine( Const Line: String ): Boolean;
begin
   Result := (Line>'') and (Line[1] in ['=','a'..'z','A'..'Z'])
end;

procedure TFiltersMail.SelectSections( Const SectionIdentifier: String );
var  LineNo: Integer;
     Line  : TFilterLineMail;
begin
     inherited SelectSections( SectionIdentifier );
     TOP_makes_sense := False;
     for LineNo:=0 to LinesCount-1 do begin
        Line := TFilterLineMail( LinesItem[ LineNo ] );
        if Line.ActionID in [MLFLT_ACTION_KILL, MLFLT_ACTION_IGNORE] then begin
           TOP_makes_sense := True;
           break;
        end;
     end;
end;

procedure TFiltersMail.FilterMail( MailHeaders      : String;
                                   DefaultUser      : String;
                                   var ResultIgnore : Boolean;
                                   var ResultKillIt : Boolean;
                                   ResultNotifys    : TStrings;
                                   ResultUsers      : TStrings;
                                   ResultGroups     : TStrings;
                                   var NotifyReason : String;
                                   var ResultIMAPFolder : String);
//HSR //IMAP-Folder 08/02 (Chg)
var  LineNo, i, k  : Integer;
     Line          : TFilterLineMail;
     Parser        : TParser;
     ResAccounts, s: String;
     WantNotify    : Boolean;
     Hdr           : TArticle; 
begin
   ResultIMAPFolder := ''; //HSR //IMAP-Folder 08/02
   ResultIgnore := False;
   ResultKillIt := False;
   WantNotify   := False;
   ResultNotifys.Clear;
   ResultUsers.Clear;
   ResultGroups.Clear;
   NotifyReason := '';
   ResultIMAPFolder := '';

   Parser := TParser.Create;
   try
     {JW} {Postmaster}
     if DefaultUser='' then DefaultUser:=Def_Postmaster;
//     if DefaultUser='' then DefaultUser:='admin';
     {/JW}
     i := Pos( #13#10#13#10, MailHeaders );
     if i>0 then MailHeaders := copy( MailHeaders, 1, i+1 );

     Hdr := TArticle.Create;
     try
        Hdr.Text := MailHeaders;
        for LineNo:=0 to LinesCount-1 do begin
           Line := TFilterLineMail( LinesItem[ LineNo ] );
           if Line.MatchesMailHeaders( RegexFilter, Hdr, ResAccounts ) then begin
              Log( LOGID_DEBUG, 'MailFilter (+): ' + Line.AsString );
              case Line.ActionID of

                 // undo a previous kill/ignore
                 MLFLT_ACTION_LOAD: begin
                    ResultIgnore := False;
                    ResultKillIt := False;
                    NotifyReason := '';
                 end;

                 // don't load, but delete
                 MLFLT_ACTION_KILL: begin
                    ResultKillIt := True;
                    NotifyReason := NotifyReason + '-> ' + Line.AsString + #13#10;
                 end;

                 // don't load, don't delete
                 MLFLT_ACTION_IGNORE: begin
                    ResultKillIt := False;
                    ResultIgnore := True;
                    NotifyReason := NotifyReason + '-> ' + Line.AsString + #13#10;
                 end;

                 // notify if "killed" or "ignored"
                 MLFLT_ACTION_NOTIFY: begin
                    WantNotify := True;
                    ResultNotifys.Clear;
                    Parser.Parse( Line.ActionPars, ',' );
                    for i:=0 to Parser.Count-1 do begin
                       s := TrimWhSpace( Parser.sPart( i, '' ) );
                       if s<>'' then ResultNotifys.Add( s );
                    end;
                 end;

                 // do NOT notify anyone
                 MLFLT_ACTION_NOTIFYOFF: begin   // JAWO 21.09.01 (Notify none)
                    WantNotify := false;
                    ResultNotifys.Clear;
                 end;

                 // new default-user
                 MLFLT_ACTION_DEFAULT: begin
                    DefaultUser := Line.ActionPars;
                 end;

                 // set/start new recipient-list
                 MLFLT_ACTION_SET: begin
                    ResultUsers.Clear;
                    Parser.Parse( Line.ActionPars, ',' );
                    for i:=0 to Parser.Count-1 do begin
                       s := TrimWhSpace( Parser.sPart( i, '' ) );
                       if s<>'' then ResultUsers.Add( s );
                    end;
                 end;

                 // add to recipient-list
                 MLFLT_ACTION_ADD: begin
                    Parser.Parse( Line.ActionPars, ',' );
                    for i:=0 to Parser.Count-1 do begin
                       s := TrimWhSpace( Parser.sPart( i, '' ) );
                       k := ResultUsers.IndexOf( s );
                       if (k<0) and (s<>'') then ResultUsers.Add( s );
                    end;
                 end;

                 // delete from recipient-list
                 MLFLT_ACTION_DEL: begin
                    Parser.Parse( Line.ActionPars, ',' );
                    for i:=0 to Parser.Count-1 do begin
                       s := TrimWhSpace( Parser.sPart( i, '' ) );
                       k := ResultUsers.IndexOf( s );
                       if k>=0 then ResultUsers.Delete( k );
                    end;
                 end;

                 // add matching accounts to recipient-list
                 MLFLT_ACTION_ADDACCOUNTS: begin
                    Parser.Parse( ResAccounts, ',' );
                    for i:=0 to Parser.Count-1 do begin
                       s := TrimWhSpace( Parser.sPart( i, '' ) );
                       if s<>'' then ResultUsers.Add( s );
                    end;
                 end;

                 // add matching accounts to recipient-list
                 MLFLT_ACTION_ADDDEFAULT: begin
                    ResultUsers.Add( DefaultUser  );
                 end;

                 // post to newsgroups
                 MLFLT_ACTION_POSTTO: begin
                    Parser.Parse( Line.ActionPars, ',' );
                    for i:=0 to Parser.Count-1 do begin
                       s := TrimWhSpace( Parser.sPart( i, '' ) );
                       if s<>'' then ResultGroups.Add( s );
                    end;
                 end;

                 //IMAP-Folder
                 MLFLT_ACTION_FOLDER: begin //HSR //IMAP-Folder 08/02
                   ResultIMAPFolder := Line.ActionPars;
                 end;

              end;

              if Line.IsFinal then break; // line preceded with "="

           end else begin
              {Log( LOGID_FULL, 'MailFilter (-): ' + Line.AsString );}
           end
        end
     finally
        Hdr.free
     end;

     // use default-user, if no recipients/newsgroups were selected
     if (ResultUsers.Count=0) and (ResultGroups.Count=0) then begin
        if DefaultUser<>'' then begin
           Parser.Parse( DefaultUser, ',' );
           for i:=0 to Parser.Count-1 do begin
              s := TrimWhSpace( Parser.sPart( i, '' ) );
              if s<>'' then ResultUsers.Add( s );
           end;
        end;
     end;

     // fill list of notification-recipients if none were given with "notify()"
     if WantNotify and (ResultIgnore or ResultKillIt) then begin
        // send notifications to determined recipients ...
        if ResultNotifys.Count=0 then ResultNotifys.Text:=ResultUsers.Text;
        // .. or finally to "admin"
        {JW} {Postmaster}
        // (now to Postmaster!)
        if ResultNotifys.Count=0 then ResultNotifys.Add( Def_Postmaster );
//        if ResultNotifys.Count=0 then ResultNotifys.Add( 'admin' );
        {/JW}
     end;
   finally
     Parser.Free
   end
end;

procedure TFiltersMail.Purge;
Var ok: Boolean; LineNo: Integer; LineText: String;
begin
   ok := false;
   for LineNo:=fFilterFile.Count-1 downto 0 do begin
      LineText := TrimWhSpace( fFilterFile[LineNo] );
      if LineText > '' then If IsFilterLine( LineText ) then begin
         With TFilterLineMail.Create do try
            If SetFilterLine( LineText ) then begin
               If (Expire > 0) and (Expire <= Now) then begin
                  If Def_Mail_ExpiredFilterentries_Delete then begin
                     fFilterFile.Delete (LineNo);
                     LogFile.Add ( LOGID_INFO, TrGlF(kLog, 'Mailfilterfile.ExpiredEntry.Deleted',
                        'Expired Mail-filter-file entry deleted: "%s"', LineText))
                  end else begin
                     fFilterFile[LineNo] := '# Expired! # '+LineText;
                     LogFile.Add ( LOGID_INFO, TrGlF(kLog, 'Mailfilterfile.ExpiredEntry.commented',
                        'Expired Mail-filter-file entry commented out: "%s"', LineText))
                  end;
                  ok := true
               end
   end
         finally Free end
      end
   end;
   If ok then fFilterFile.SaveToFile ( fFilterFilename );
end;

procedure TFiltersMail.Test;
var  LineNo, Counter: Integer;
     Line  : TFilterLineMail;
     LineText: String;
begin
   If fFilterFile.Count = 0 then Exit;
   Log( LOGID_SYSTEM, TrGlF(kLog, 'Mailfilterfile.Testing',
      'Mail-filter-file %s: Testing', fFilterFilename ));
   Line := TFilterLineMail.Create;
   Counter := 0;
   try
      for LineNo:=0 to fFilterFile.Count-1 do begin
         LineText := TrimWhSpace( fFilterFile[LineNo] );
         if IsFilterLine( LineText ) then begin
            if Not Line.SetFilterLine( LineText ) then begin
               Log( LOGID_WARN, TrGlF(kLog, 'Mailfilterfile.LineError',
                  'Error in line %s of Mail-filter-file: %s', [IntToStr(LineNo+1), LineText] ));
               Log( LOGID_WARN, TrGl(kLog, 'Mailfilterfile.ErrorDescr',
                  'Error description: ') + line.LastSetError );
               Inc(Counter)
            end
         end
      end
   finally
      Line.free
   end;
   If Counter = 0
      then Log( LOGID_SYSTEM, TrGlF(kLog, 'Mailfilterfile.TestOK',
         'Mail-filter-file %s: Test OK', fFilterFilename ))
      else Log( LOGID_SYSTEM, TrGlF(kLog, 'Mailfilterfile.ErrorsFound',
         'Mail-filter-file %s: %s errors found', [fFilterFilename, IntToStr(Counter)] ));
end;

end.
