// ============================================================================
// "Input"-dialog (string/password/integer-input, list-selection)
// Copyright (c) 1999, Juergen Haible. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
// ============================================================================

unit dInput; // "Input"-dialog (string/password/integer-input, list-selection)

// ----------------------------------------------------------------------------
// Contains an input-dialog which either allows input of a single value
// (string, password or integer) or the selection of a list-entry.
// ----------------------------------------------------------------------------

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
  Buttons, ExtCtrls, cStdForm, Messages;

type
  TInputDlgType = ( idtString, idtPassword, idtInteger, idtList );

  TInputDlg = class(THForm)
    Panel1: TPanel;
    OKBtn: TButton;
    CancelBtn: TButton;
    Panel2: TPanel;
    lblPrompt: TLabel;
    lbSelect: TListBox;
    emInput: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure lbSelectDblClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    DlgType: TInputDlgType;
    IntMin, IntMax: LongInt;
    procedure PrepareForm( ACaption, APrompt, AValue: String;
                           AType: TInputDlgType;
                           AList: String;
                           ATop, ALeft, AHeight, AWidth: Integer );
  end;

function InputDlgStr( const ACaption, APrompt: String;
                      var Value: String; HelpID, ATop, ALeft: Integer ): Boolean;

function InputDlgPwd( const ACaption, APrompt: String;
                      var Value: String; HelpID, ATop, ALeft: Integer ): Boolean;

function InputDlgInt( const ACaption, APrompt: String;
                      var Value: LongInt;
                      AIntMin, AIntMax: LongInt; HelpID: Integer ): Boolean;

function InputDlgList( const ACaption, APrompt: String;
                       var Value: LongInt;
                       const AList: String; HelpID, ATop, ALeft, AHeight, AWidth: Integer ): Boolean;

// ----------------------------------------------------------------------------

implementation

Uses Config;

{$R *.DFM}

// ------------------------------------------------------ TInputDlgThread -----

type
  TInputDlgThread = class( TThread )
    private
       FCaption, FPrompt, FValue, FList: String;
       FIntMin, FIntMax: LongInt;
       FType: TInputDlgType;
       FResult: Boolean;
       FTop, FLeft, FHeight, FWidth: Integer;
       procedure SyncDialog;
    public
       property Result: Boolean read FResult;
       property Value : String  read FValue;
       procedure Execute; override;
       constructor Create( const ACaption, APrompt, AValue: String;
                           const AType: TInputDlgType;
                           const AIntMin, AIntMax: LongInt;
                           const AList: String;
                           const ATop, ALeft, AHeight, AWidth: Integer );
  end;

function ExecuteDialog( const ACaption, APrompt: String;
                        var   AValue: String;
                        const AType: TInputDlgType;
                        const AIntMin, AIntMax: LongInt;
                        const AList: String;
                        const ATop, ALeft, AHeight, AWidth: Integer ): Boolean;
begin
   Result := False;
   With TInputDlg.Create(Application) do try
      PrepareForm( ACaption, APrompt, AValue, AType, AList, ATop, ALeft, AHeight, AWidth );
      IntMin := AIntMin;
      IntMax := AIntMax;
      If ShowModal=mrOK then begin
         if AType=idtList
            then AValue := IntToStr(lbSelect.ItemIndex)
            else AValue := emInput.Text;
         Result := True
      end
   finally
      release
   end
end;

procedure TInputDlgThread.SyncDialog;
begin
   FResult := ExecuteDialog( FCaption, FPrompt, FValue, FType, FIntMin,
                             FIntMax, FList, FTop, FLeft, FHeight, FWidth )
  end;

procedure TInputDlgThread.Execute;
begin
   try Synchronize( SyncDialog ) finally Terminate end
end;

constructor TInputDlgThread.Create( const ACaption, APrompt, AValue: String;
                                    const AType: TInputDlgType;
                                    const AIntMin, AIntMax: LongInt;
                                    const AList: String;
                                    const ATop, ALeft, AHeight, AWidth: Integer );
begin
   inherited Create( True );
   FCaption := ACaption;
   FPrompt  := APrompt;
   FValue   := AValue;
   FType    := AType;
   FIntMin  := AIntMin;
   FIntMax  := AIntMax;
   FList    := AList;
   FTop     := ATop;
   FLeft    := ALeft;
   FHeight  := AHeight;
   FWidth   := AWidth;
   Resume;
end;

// ------------------------------------------------------------ functions -----

function InputDlgSyncVCL( const ACaption, APrompt: String;
                          var Value: String;
                          AType: TInputDlgType;
                          AIntMin, AIntMax: LongInt;
                          const AList: String;
                          ATop, ALeft, AHeight, AWidth: Integer ): Boolean;
var  ThdDlg: TInputDlgThread;
begin
   if GetCurrentThreadID = MainThreadID then begin
      // running in VCL-thread, no need to synchronize
      Result := ExecuteDialog( ACaption, APrompt, Value, AType, AIntMin,
                               AIntMax, AList, ATop, ALeft, AHeight, AWidth )
   end else begin
      // we need to synchronize dialog with VCL-thread
      ThdDlg := TInputDlgThread.Create( ACaption, APrompt, Value, AType, AIntMin,
                                        AIntMax, AList, ATop, ALeft, AHeight, AWidth );
      try
         ThdDlg.WaitFor;
         Result := ThdDlg.Result;
         Value  := ThdDlg.Value;
      finally
         ThdDlg.Free;
      end
   end
end;

function InputDlgStr( const ACaption, APrompt: String;
                      var Value: String; HelpID, ATop, ALeft: Integer ): Boolean;
begin
   Result := InputDlgSyncVCL( ACaption, APrompt, Value, idtString,
                              -1, -1, '', ATop, ALeft, -1, -1 );
end;

function InputDlgPwd( const ACaption, APrompt: String;
                      var Value: String; HelpID, ATop, ALeft: Integer ): Boolean;
begin
   Result := InputDlgSyncVCL( ACaption, APrompt, Value, idtPassword,
                              -1, -1, '', ATop, ALeft, -1, -1 );
end;

function InputDlgInt( const ACaption, APrompt: String;
                      var Value: LongInt;
                      AIntMin, AIntMax: LongInt; HelpID: Integer ): Boolean;
var  sValue: String;
begin
   sValue := inttostr( Value );
   Result := InputDlgSyncVCL( ACaption, APrompt, sValue,
                              idtInteger, AIntMin, AIntMax, '', -1, -1, -1 ,-1 );
   if Result then Value := StrToIntDef( sValue, Value );
end;

function InputDlgList( const ACaption, APrompt: String;
                       var Value: LongInt;
                       const AList: String;
                       HelpID, ATop, ALeft, AHeight, AWidth: Integer): Boolean;
var  sValue: String;
begin
   sValue := inttostr( Value );
   Result := InputDlgSyncVCL( ACaption, APrompt, sValue,
                              idtList, -1, -1, AList, ATop, ALeft, AHeight, AWidth );
   if Result then Value := StrToIntDef( sValue, Value );
end;

// ------------------------------------------------------------ TInputDlg -----

procedure TInputDlg.PrepareForm( ACaption, APrompt, AValue: String;
                                 AType: TInputDlgType; AList: String;
                                 ATop, ALeft, AHeight, AWidth: Integer );
var  MX, CH, TH, TW, i, p, TempLen, MaxLen: Integer;
     s, NPrompt: String;
     X: TWinControl;

begin
     DlgType := AType;

     MX := lblPrompt.ClientWidth;
     CH := Canvas.TextHeight( 'Gg' );
     TH := 2;
     NPrompt := '';

     while APrompt<>'' do begin
        i := Pos( #13#10, APrompt );
        if i=0 then begin
           s := APrompt;
           APrompt := '';
        end else begin
           s := copy( APrompt, 1, i-1 );
           System.Delete( APrompt, 1, i+1 );
        end;

        if Canvas.TextWidth(s)<=MX then begin
           NPrompt := NPrompt + s + #13#10;
           inc( TH, CH );
        end else begin
           repeat
              p := 0;
              for i:=1 to length(s) do begin
                 /// #TGL#
                 if (s[i]=' ') {(Pos(s[i],' -,.;:!?' )>0)} or (i=length(s)) then begin
                    TW := Canvas.TextWidth( copy(s,1,i-1) );
                    if TW>MX then begin
                       if p=0 then p:=i-1;
                       break;
                    end;
                    if TW<=MX then p:=i;
                 end;
              end;
              if p<=0 then p:=length(s);

              NPrompt := NPrompt + copy( s, 1, p ) + #13#10;
              inc( TH, CH );
              System.Delete( s, 1, p );
              if copy(s,1,1)=' ' then System.Delete( s, 1, 1 );
           until s='';
        end;
     end;

     Caption           := ACaption;
     lblPrompt.Height  := TH;
     lblPrompt.Caption := NPrompt;

     if DlgType=idtList then begin
        X := lbSelect;
        emInput.Visible  := False;
        lbSelect.Visible := True;
        lbSelect.Items.Text := AList;
        if lbSelect.Items.Count>0 then begin
           i := 0;
           MaxLen := 0;
           p := strtointdef( AValue, 0 );
           if (p>=0) and (p<lbSelect.Items.Count) then i:=p;
           lbSelect.ItemIndex := i;
           For i := 0 To lbSelect.Items.count - 1 Do
           Begin
              TempLen := lbSelect.Canvas.TextWidth(lbSelect.Items[I]);
              If TempLen > MaxLen Then MaxLen := TempLen;
           End;
           sendmessage(lbSelect.Handle, LB_SETHORIZONTALEXTENT, MaxLen + 10, 0);
        end;
     end else begin
        X := emInput;
        X.Height := CH * 2 + 22;
        emInput.Text := AValue;
     end;

     X.Top := lblPrompt.Top + lblPrompt.Height + 4;
     ClientHeight :=  X.Height + lblPrompt.Height + 18;
     If AHeight <> -1 then ClientHeight := AHeight;
     If AWidth <> -1 then ClientWidth := AWidth;
     if ATop = -1 then Top := (Screen.Height - Height) div 2 else Top := ATop;
     if ALeft = -1 then Left := (Screen.Width - Width) div 2 else Left := ALeft;

     if DlgType=idtPassword then emInput.PasswordChar:='*';
end;

procedure TInputDlg.FormCreate(Sender: TObject);
begin
   DlgType := idtString;
   IntMin := LongInt(MINLONG);
   IntMax := LongInt(MAXLONG);
end;

procedure TInputDlg.OKBtnClick(Sender: TObject);
Var l: LongInt;
begin
   case DlgType of
      idtInteger:
         try
            l := SysUtils.StrToInt(emInput.Text);
            if l<IntMin then ModalResult:=mrNone;
            if l>IntMax then ModalResult:=mrNone;
         except ModalResult:=mrNone end;
   end;
   if ModalResult=mrNone then begin
      emInput.Color := clRed;
      emInput.SetFocus;
   end;
end;

// ----------------------------------------------------------------------------

procedure TInputDlg.lbSelectDblClick(Sender: TObject);
begin
   ModalResult := mrOK
end;

end.
